/*
 *  trender - 3D Studio .ASC files realtime renderer
 *  Copyright (C) 1999  Thanassis Tsiodras (ttsiod@softlab.ntua.gr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "Vgalib.h"
#include "3DTypes.h"

#define DIFFUSE 128
#define AMBIENT 32
#define SPEC_COEFF 8.0
#define SPECULAR

Vertex::Vertex(Object3D *object, FLOAT x, FLOAT y, FLOAT z)
{
    usX = x; usY = y; usZ = z;

    fNormalX = fNormalY = fNormalZ = 0;
    fTransformedX = fTransformedY = fTransformedZ = 0;
    always_draw = FALSE;
    intensity = 0;

    next = object->GetHeadVertex();
    object->SetHeadVertex(this);
}

void Vertex::Illuminate(Light *light, Camera *camera)
{
    FLOAT fSumm;
    FLOAT lx,ly,lz;   /* light vector */
    FLOAT rx,ry,rz;   /*  reflection  */
    FLOAT fSpecularIntensity = 0.0;

/* Diffuse light */
    lx = light->GetX() - usX;
    ly = light->GetY() - usY;
    lz = light->GetZ() - usZ;

    fSumm=sqrt(lx*lx + ly*ly + lz*lz);
    if (fSumm<1e-5) {
        intensity = AMBIENT;
	return;
    }

    /* normalize light vector */
    lx/=fSumm;
    ly/=fSumm;
    lz/=fSumm;

    /* reflection vector: r = -light + 2*normal*cos phi,
       where phi the angle between normal and light

       first, r = -light (normalized) */

    rx = -lx;
    ry = -ly;
    rz = -lz;

    /* Diffuse intensity = normalized light . normalized normal
       Ranges from 0..DIFFUSE */
    intensity = (LONG) (DIFFUSE*(lx*fNormalX +
			         ly*fNormalY +
			         lz*fNormalZ));

/* check if vertex is looking away from the light... */
    if (intensity<=0) {
	intensity=AMBIENT; // ambient intensity
	return;
    }
     
/* Specular light */

    /* First, complete the reflection calculation:

       reflection vector: r = -light + 2*normal*cos phi,
       where phi the angle between normal and light.

       We have just calculated diffuse intensity as DIFFUSE*(normal . light)
       so, cos(phi) = intensity/(DIFFUSE*(|normal| |light|). And since normal
       and light are normalized, ... */

#ifdef SPECULAR
    rx += 2.0*intensity*fNormalX/DIFFUSE;
    ry += 2.0*intensity*fNormalY/DIFFUSE;
    rz += 2.0*intensity*fNormalZ/DIFFUSE;

    FLOAT ex, ey, ez; // eye vector

    ex = camera->GetEyeX() - usX;
    ey = camera->GetEyeY() - usY;
    ez = camera->GetEyeZ() - usZ;

    // normalize eye vector 
    fSumm=sqrt(ex*ex + ey*ey + ez*ez);
    if (fSumm<1e-5) {
	// intensity: only ambient and diffuse
	goto ambient;
    }
    ex = ex/fSumm; ey = ey/fSumm; ez = ez/fSumm;

    // normalize reflection vector
    fSumm=sqrt(rx*rx + ry*ry + rz*rz);
    if (fSumm<1e-5) {
	// intensity: only ambient and diffuse
	goto ambient;
    }
    rx = rx/fSumm; ry = ry/fSumm; rz = rz/fSumm;

    // Calculation of specular intensity
    fSpecularIntensity = (ex*rx + ey*ry + ez*rz);
    if (fSpecularIntensity < 0) {
	// intensity: only ambient and diffuse
	goto ambient;
    }

    // The normal way to calculate specular intensity is to use 
    // the following equation:
    //
    // fSpecularIntensity = pow(fSpecularIntensity, SPEC_COEFF);
    //
    // However, this is a SLOW calculation. For a SPEC_COEFF of 8,
    // a much faster way to do it is:

    fSpecularIntensity*=fSpecularIntensity;
    fSpecularIntensity*=fSpecularIntensity;
    fSpecularIntensity*=fSpecularIntensity;

    // specularIntensity: 0..73
    intensity += (LONG)(fSpecularIntensity*(255-AMBIENT-DIFFUSE));
#endif

/* Ambient */
ambient:
    intensity+=AMBIENT;
    if (intensity>255)intensity=255;
    if (intensity<0)intensity=0;
}

void Vertex::Transform(Camera *camera)
{
    fTransformedX = camera->m00*usX +
		    camera->m01*usY +
		    camera->m02*usZ +
		    camera->m03;

    fTransformedY = camera->m10*usX +
		    camera->m11*usY +
		    camera->m12*usZ +
		    camera->m13;

    fTransformedZ = camera->m20*usX +
		    camera->m21*usY +
		    camera->m22*usZ +
		    camera->m23;

}
