/*
 *  trender - 3D Studio .ASC files realtime renderer
 *  Copyright (C) 1999  Thanassis Tsiodras (ttsiod@softlab.ntua.gr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include "Wu.h"
#include "Defines.h"

Uint8 clipper[256*3];
Uint32 offsets[MAXY];
Uint32 offset;

void
Initialize()
{
        for(int i=0; i<256; i++)
                clipper[i]=0;
        for(int i=0; i<256; i++)
                clipper[i+256]=i;
        for(int i=0; i<256; i++)
                clipper[i+512]=255;

        for(int i=0; i<MAXY; i++)
                offsets[i] = MAXX*i;
}

/* Wu antialiased line drawer.
 * (X0,Y0),(X1,Y1) = line to draw
 * BaseColor = color # of first color in block used for antialiasing, the
 *          0% intensity version of the drawing color
 * NumLevels = size of color block, with BaseColor+NumLevels-1 being the
 *          100% intensity version of the drawing color
 * IntensityBits = log base 2 of NumLevels; the # of bits used to describe
 *          the intensity of the drawing color. 2**IntensityBits==NumLevels
 */
void DrawWuLine(
    Uint8 *pixelsoriginal,
    int X0,
    int Y0,
    int X1,
    int Y1,
    int BaseColor,
    int NumLevels,
    unsigned short IntensityBits)
{
    unsigned short IntensityShift, ErrorAdj, ErrorAcc;
    unsigned short ErrorAccTemp, Weighting, WeightingComplementMask;
    int DeltaX, DeltaY, Temp, XDir;
    Uint8 *pixels = pixelsoriginal;

    /* Make sure the line runs top to bottom */
    if (Y0 > Y1) {
	Temp = Y0; Y0 = Y1; Y1 = Temp;
	Temp = X0; X0 = X1; X1 = Temp;
    }
    /* Draw the initial pixel, which is always exactly intersected by
       the line and so needs no weighting */

    pixels += offsets[Y0] + X0;
    *pixels = clipper[256 + *pixels + BaseColor+NumLevels-1];

    if ((DeltaX = X1 - X0) >= 0) {
	XDir = 1;
    } else {
	XDir = -1;
	DeltaX = -DeltaX; /* make DeltaX positive */
    }
    
    /* Special-case horizontal, vertical, and diagonal lines, which
       require no weighting because they go right through the center of
       every pixel */

    if ((DeltaY = Y1 - Y0) == 0) {
	/* Horizontal line */
	while (DeltaX-- != 0) {
	    pixels += XDir;
	    *pixels = clipper[256 + *pixels + BaseColor+NumLevels-1];
	}
	return;
    }
    if (DeltaX == 0) {
	/* Vertical line */
	do {
	    pixels += MAXX;
	    *pixels = clipper[256 + *pixels + BaseColor+NumLevels-1];
	} while (--DeltaY != 0);
	return;
    }
    if (DeltaX == DeltaY) {
	/* Diagonal line */
	do {
	    pixels += XDir;
	    pixels += MAXX;
	    *pixels = clipper[256 + *pixels + BaseColor+NumLevels-1];
	} while (--DeltaY != 0);
	return;
    }

    /* Line is not horizontal, diagonal, or vertical */

    ErrorAcc = 0;  /* initialize the line error accumulator to 0 */

    /* # of bits by which to shift ErrorAcc to get intensity level */
    IntensityShift = 16 - IntensityBits;

    /* Mask used to flip all bits in an intensity weighting, producing the
       result (1 - intensity weighting) */
    WeightingComplementMask = NumLevels - 1;

    /* Is this an X-major or Y-major line? */
    if (DeltaY > DeltaX) {
	/* Y-major line; calculate 16-bit fixed-point fractional part of a
	   pixel that X advances each time Y advances 1 pixel, truncating the
	   result so that we won't overrun the endpoint along the X axis */
	ErrorAdj = ((unsigned long) DeltaX << 16) / (unsigned long) DeltaY;

	/* Draw all pixels other than the first and last */
	while (--DeltaY) {
	    ErrorAccTemp = ErrorAcc;   /* remember currrent accumulated error */
	    ErrorAcc += ErrorAdj;      /* calculate error for next pixel */
	    if (ErrorAcc <= ErrorAccTemp) {
		/* The error accumulator turned over, so advance the X coord */
		pixels += XDir;
	    }
	    pixels += MAXX; /* Y-major, so always advance Y */

	    /* The IntensityBits most significant bits of ErrorAcc give us the
	       intensity weighting for this pixel, and the complement of the
	       weighting for the paired pixel */
	    Weighting = ErrorAcc >> IntensityShift;
	    
	    *pixels = clipper[256 + *pixels + 
	    	BaseColor + (Weighting ^ WeightingComplementMask)];
	    *(pixels+XDir) = clipper[256 + *(pixels+XDir) + 
	    	BaseColor + Weighting];
	}
	/* Draw the final pixel, which is always exactly intersected by the line
	   and so needs no weighting */
	pixelsoriginal += offsets[Y1] + X1;
	*pixelsoriginal= clipper[256+ *pixelsoriginal + BaseColor+NumLevels-1];
	return;
    }
    /* It's an X-major line; calculate 16-bit fixed-point fractional part of a
       pixel that Y advances each time X advances 1 pixel, truncating the
       result to avoid overrunning the endpoint along the X axis */
    ErrorAdj = ((unsigned long) DeltaY << 16) / (unsigned long) DeltaX;

    /* Draw all pixels other than the first and last */
    while (--DeltaX) {
	ErrorAccTemp = ErrorAcc;   /* remember currrent accumulated error */
	ErrorAcc += ErrorAdj;      /* calculate error for next pixel */
	if (ErrorAcc <= ErrorAccTemp) {
	    /* The error accumulator turned over, so advance the Y coord */
	    pixels += MAXX;
	}
	pixels += XDir; /* X-major, so always advance X */
	/* The IntensityBits most significant bits of ErrorAcc give us the
	   intensity weighting for this pixel, and the complement of the
	   weighting for the paired pixel */
	Weighting = ErrorAcc >> IntensityShift;

	*pixels = clipper[256 + *pixels + 
	    BaseColor + (Weighting ^ WeightingComplementMask)];
	*(pixels+MAXX) = clipper[256 + *(pixels+MAXX) + 
	    BaseColor + Weighting];
    }
    /* Draw the final pixel, which is always exactly intersected by the line
       and so needs no weighting */
    pixelsoriginal += offsets[Y1] + X1;
    *pixelsoriginal= clipper[256+ *pixelsoriginal + BaseColor+NumLevels-1];
}
